@@ -458,16 +458,18 @@ void InstanceView::paintEvent([[maybe_unused]] QPaintEvent* event)
     QPainter painter(this->viewport());
 
     if (m_catVisible) {
+        painter.setOpacity(APPLICATION->settings()->get("CatOpacity").toFloat() / 100);
         int widWidth = this->viewport()->width();
         int widHeight = this->viewport()->height();
         if (m_catPixmap.width() < widWidth)
             widWidth = m_catPixmap.width();
         if (m_catPixmap.height() < widHeight)
             widHeight = m_catPixmap.height();
-        auto pixmap = m_catPixmap.scaled(widWidth, widHeight, Qt::KeepAspectRatio);
+        auto pixmap = m_catPixmap.scaled(widWidth, widHeight, Qt::KeepAspectRatio, Qt::SmoothTransformation);
         QRect rectOfPixmap = pixmap.rect();
         rectOfPixmap.moveBottomRight(this->viewport()->rect().bottomRight());
         painter.drawPixmap(rectOfPixmap.topLeft(), pixmap);
+        painter.setOpacity(1.0);
     }
 
 #if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
@@ -846,7 +848,7 @@ QRegion InstanceView::visualRegionForSelection(const QItemSelection& selection)
     return region;
 }
 
-QModelIndex InstanceView::moveCursor(QAbstractItemView::CursorAction cursorAction, [[maybe_unused]] Qt::KeyboardModifiers modifiers)
+QModelIndex InstanceView::moveCursor(QAbstractItemView::CursorAction cursorAction, Qt::KeyboardModifiers modifiers)
 {
     auto current = currentIndex();
     if (!current.isValid()) {
@@ -863,6 +865,7 @@ QModelIndex InstanceView::moveCursor(QAbstractItemView::CursorAction cursorActio
     if (m_currentCursorColumn < 0) {
         m_currentCursorColumn = column;
     }
+    // Handle different movement actions.
     switch (cursorAction) {
         case MoveUp: {
             if (row == 0) {
@@ -923,16 +926,47 @@ QModelIndex InstanceView::moveCursor(QAbstractItemView::CursorAction cursorActio
             if (column > 0) {
                 m_currentCursorColumn = column - 1;
                 return cat->rows[row][column - 1];
+            } else if (row > 0) {
+                row -= 1;
+                int newRowSize = cat->rows[row].size();
+                m_currentCursorColumn = newRowSize - 1;
+                return cat->rows[row][m_currentCursorColumn];
+            } else {
+                int prevGroupIndex = group_index - 1;
+                while (prevGroupIndex >= 0) {
+                    auto prevGroup = m_groups[prevGroupIndex];
+                    if (prevGroup->collapsed) {
+                        prevGroupIndex--;
+                        continue;
+                    }
+                    int lastRow = prevGroup->numRows() - 1;
+                    int lastCol = prevGroup->rows[lastRow].size() - 1;
+                    m_currentCursorColumn = lastCol;
+                    return prevGroup->rows[lastRow][lastCol];
+                }
             }
-            // TODO: moving to previous line
             return current;
         }
         case MoveRight: {
             if (column < cat->rows[row].size() - 1) {
                 m_currentCursorColumn = column + 1;
                 return cat->rows[row][column + 1];
+            } else if (row < cat->rows.size() - 1) {
+                row += 1;
+                m_currentCursorColumn = 0;
+                return cat->rows[row][m_currentCursorColumn];
+            } else {
+                int nextGroupIndex = group_index + 1;
+                while (nextGroupIndex < m_groups.size()) {
+                    auto nextGroup = m_groups[nextGroupIndex];
+                    if (nextGroup->collapsed) {
+                        nextGroupIndex++;
+                        continue;
+                    }
+                    m_currentCursorColumn = 0;
+                    return nextGroup->rows[0][0];
+                }
             }
-            // TODO: moving to next line
             return current;
         }
         case MoveHome: {
@@ -945,6 +979,7 @@ QModelIndex InstanceView::moveCursor(QAbstractItemView::CursorAction cursorActio
             return cat->rows[row][last];
         }
         default:
+            // For unsupported cursor actions, return the current index.
             break;
     }
     return current;
