@@ -6,7 +6,8 @@
 enum UserDataTypes {
     TITLE = 257,        // QString
     DESCRIPTION = 258,  // QString
-    SELECTED = 259      // bool
+    SELECTED = 259,     // bool
+    INSTALLED = 260     // bool
 };
 
 /** This is an item delegate composed of:
@@ -17,9 +18,8 @@ enum UserDataTypes {
 class ProjectItemDelegate final : public QStyledItemDelegate {
     Q_OBJECT
 
-    public:
-        ProjectItemDelegate(QWidget* parent);
-
-        void paint(QPainter*, const QStyleOptionViewItem&, const QModelIndex&) const override;
+   public:
+    ProjectItemDelegate(QWidget* parent);
 
+    void paint(QPainter*, const QStyleOptionViewItem&, const QModelIndex&) const override;
 };
